"""Python Smarty."""

from .registers import Registers
from .connection import Connection


class Smarty:
    """Smarty Class."""

    def __init__(self, host, port=502, slave=1, loop=None):
        self.connection = Connection(host, port, slave, loop)
        self._registers = Registers(self.connection)

    @property
    def fan_speed(self) -> int:
        """Get Current Fan Speed."""
        register = self._registers.get_register(
            'HR_USER_CONFIG_CURRENT_SYSTEM_MODE')
        return register.state

    @property
    def fan_speed_name(self) -> str:
        """Get Current Fan Speed Name."""
        register = self._registers.get_register(
            'HR_USER_CONFIG_CURRENT_SYSTEM_MODE')
        return register.state_name

    @property
    def system_state(self) -> str:
        """Get Current System State."""
        register = self._registers.get_register(
            'IR_CURRENT_SYSTEM_STATE')
        return register.state_name

    @property
    def filter_timer(self) -> int:
        """Get Filter Timer Days Left."""
        register = self._registers.get_register(
            'IR_FILTERS_TIMER_DAYS_LEFT')
        return register.state

    @property
    def supply_fan_speed(self) -> int:
        """Get Supply Fan Speed (RPM)."""
        register = self._registers.get_register(
            'IR_SUPPLY_FAN_SPEED_RPM')
        return register.value

    @property
    def extract_fan_speed(self) -> int:
        """Get Extract Fan Speed (RPM)."""
        register = self._registers.get_register(
            'IR_EXTRACT_FAN_SPEED_RPM')
        return register.value

    @property
    def supply_air_temperature(self) -> int:
        """Get Supply Air Temperature."""
        register = self._registers.get_register(
            'IR_SUPPLY_AIR_TEMPERATURE')
        return register.value

    @property
    def extract_air_temperature(self) -> int:
        """Get Extract Air Temperature."""
        register = self._registers.get_register(
            'IR_EXTRACT_AIR_TEMPERATURE')
        return register.value

    @property
    def outdoor_air_temperature(self) -> int:
        """Get Outdoor Air Temperature."""
        register = self._registers.get_register(
            'IR_OUTDOOR_AIR_TEMPERATURE')
        return register.value

    @property
    def alarm(self) -> bool:
        """Get Alarm."""
        register = self._registers.get_register(
            'HR_ALARM_A')
        return bool(register.state)

    @property
    def warning(self) -> bool:
        """Get Warning."""
        register = self._registers.get_register(
            'HR_ALARM_B')
        return bool(register.state)

    @property
    def boost(self) -> bool:
        """Get the Boost State."""
        register = self._registers.get_register(
            'COIL_INTENSIVE_AIR_FLOW_BOOST')
        return bool(register.state)

    def get_software_version(self) -> str:
        """Software version."""
        register = self._registers.get_register(
            'IR_SOFTWARE_VERSION')
        return register.state

    def get_configuration_version(self) -> str:
        """Configuration version."""
        register = self._registers.get_register(
            'IR_CONFIGURATION_VERSION')
        return register.state

    def update(self) -> bool:
        """Update registers."""
        if self.connection.is_connected():
            return self._registers.update()
        return False

    def set_fan_speed(self, speed) -> bool:
        """Set Current Fan Speed."""
        if self.connection.is_connected():
            register = self._registers.get_register(
                'HR_USER_CONFIG_CURRENT_SYSTEM_MODE')
            register.set_state(speed)
            return True
        return False

    def turn_off(self) -> bool:
        """Turn off."""
        return self.set_fan_speed(0)

    def turn_on(self, speed) -> bool:
        """Turn on."""
        return self.set_fan_speed(speed)

    def is_on(self) -> bool:
        """Get Status."""
        return bool(self.fan_speed)

    def enable_boost(self) -> bool:
        """Set Intensive Air Flow (limited)."""
        if self.connection.is_connected():
            register = self._registers.get_register(
                'COIL_INTENSIVE_AIR_FLOW_BOOST')
            register.set_state(1)
            return True
        return False

    def disable_boost(self) -> bool:
        """Set Intensive Air Flow (limited)."""
        if self.connection.is_connected():
            register = self._registers.get_register(
                'COIL_INTENSIVE_AIR_FLOW_BOOST')
            register.set_state(0)
            return True
        return False

    def reset_filters_timer(self) -> bool:
        """Reset Filter Timer."""
        if self.connection.is_connected():
            register = self._registers.get_register(
                'COIL_FILTER_TIMER_RESET')
            register.set_state(1)
            return True
        return False



"""Discrete Input."""

from .baseregister import BaseRegister


class DiscreteInput(BaseRegister):
    """Smarty Input Register."""

    def __init__(self, **kwargs):
        super().__init__(register_type='discrete_input', **kwargs)

    def update_state(self):
        """Read Register."""
        res = self._connection.client.read_discrete_inputs(
            self.addr, slave=self._connection.slave)
        if not res.isError():
            self.state = res.bits[0]



"""Holding Register."""

from .baseregister import BaseRegister
from ..exceptions import PysmartyException


class HoldingRegister(BaseRegister):
    """Smarty Holding Register."""

    def __init__(self, **kwargs):
        super().__init__(register_type='holding_register', **kwargs)

    @property
    def state_name(self):
        """Register State Name."""
        return [k for k, v in self._states.items() if v == self.state][0]

    def update_state(self):
        """Read Register."""
        res = self._connection.client.read_holding_registers(
            self.addr, slave=self._connection.slave)
        if not res.isError():
            self.state = res.registers[0]

    def set_state(self, state):
        """Write Register."""
        if isinstance(state, int):
            if state not in self._states.values():
                raise PysmartyException('Invalid state.', state)
            value = state
        else:
            value = self._states[state]
        try:
            res = self._connection.client.write_register(
                self.addr, value, slave=self._connection.slave)
            if not res.isError():
                self.state = value
        except KeyError as ex:
            raise PysmartyException('Invalid state.', ex.args)
        except Exception as ex:
            raise PysmartyException('Something went wrong.', ex.args)



"""Smarty Modbus Registers."""

import json
from os import path

from .holding_register import HoldingRegister
from .coil import Coil
from .discrete_input import DiscreteInput
from .input_register import InputRegister

THIS_DIR = path.abspath(path.dirname(__file__))

# read/write registers
with open(path.join(THIS_DIR, "holding_registers.json"), "r") as f:
    HOLDING_REGISTERS = json.load(f)

# read/write registers
with open(path.join(THIS_DIR, "coils.json"), "r") as f:
    COILS = json.load(f)

# read only
with open(path.join(THIS_DIR, "discrete_inputs.json"), "r") as f:
    DISCRETE_INPUTS = json.load(f)

# read only
with open(path.join(THIS_DIR, "input_registers.json"), "r") as f:
    INPUT_REGISTERS = json.load(f)


class Registers:
    """
    Smarty Registers.

    MCB 1.21 Modbus Table:
    http://salda.lt/mcb/downloads/doc/MCB%201.21%20Modbus%20table%202018-05-03.xlsx
    """

    def __init__(self, conn):
        self._connection = conn
        self.holding_registers = [HoldingRegister(connection=conn, **r)
                                  for r in HOLDING_REGISTERS]
        self.coils = [Coil(connection=conn, **r) for r in COILS]
        self.discrete_inputs = [DiscreteInput(connection=conn, **r)
                                for r in DISCRETE_INPUTS]
        self.input_registers = [InputRegister(connection=conn, **r)
                                for r in INPUT_REGISTERS]
        self._registers = [self.holding_registers,
                           self.coils,
                           self.discrete_inputs,
                           self.input_registers]

    def get_register(self, register_id):
        """Get a register."""
        for regs in self._registers:
            for reg in regs:
                if reg.get_id() == register_id:
                    return reg
        return None

    def update(self) -> bool:
        """Update all registers."""
        return (self._update_holding_registers() and
                self._update_coils() and
                self._update_discrete_inputs() and
                self._update_input_registers())

    def _update_holding_registers(self) -> bool:
        """Read Holding Registers."""
        res1 = self._connection.client.read_holding_registers(
            1, 37, slave=self._connection.slave)
        res2 = self._connection.client.read_holding_registers(
            200, 3, slave=self._connection.slave)
        if res1.isError() or res2.isError():
            return False
        res = {k: v for k, v in zip(range(1, 37), res1.registers)}
        res.update({k: v for k, v in zip(range(200, 202), res2.registers)})

        update_states(res, self.holding_registers)
        return True

    def _update_coils(self) -> bool:
        """Update Coils."""
        res1 = self._connection.client.read_coils(
            1, 9, slave=self._connection.slave)
        if res1.isError():
            return False
        res = {k: v for k, v in zip(range(1, 10), res1.bits)}

        update_states(res, self.coils)
        return True

    def _update_discrete_inputs(self) -> bool:
        """Update Discrete Inputs."""
        res1 = self._connection.client.read_discrete_inputs(
            1, 67, slave=self._connection.slave)
        res2 = self._connection.client.read_discrete_inputs(
            188, 2, slave=self._connection.slave)
        if res1.isError() or res2.isError():
            return False
        res = {k: v for k, v in zip(range(1, 67), res1.bits)}
        res.update({k: v for k, v in zip(range(188, 190), res2.bits)})

        update_states(res, self.discrete_inputs)
        return True

    def _update_input_registers(self) -> bool:
        """Update input registers"""
        res1 = self._connection.client.read_input_registers(
            1, 66, slave=self._connection.slave)
        res2 = self._connection.client.read_input_registers(
            67, 66, slave=self._connection.slave)
        if res1.isError() or res2.isError():
            return False
        res = {k: v for k, v in zip(range(1, 132),
                                    res1.registers +
                                    res2.registers)}
        update_states(res, self.input_registers)
        return True


def update_states(res, registers) -> None:
    """Update Registers state."""
    for key, value in res.items():
        for reg in registers:
            if key == reg.addr:
                reg.state = value



"""Pysmarty Exceptions."""


class PysmartyBaseException(Exception):
    """Base Exception for pysmarty."""


class PysmartyException(PysmartyBaseException):
    """Pysmarty Exception."""

    def __init__(self, message, errors):
        """Exception Init."""
        super().__init__(message)
        self.errors = errors



"""Base Register."""


class BaseRegister:
    """Smarty Register."""

    def __init__(self, connection, **kwargs):
        self._connection = connection
        self._register_id = kwargs.get('ID')
        self._name = kwargs.get('NAME')
        self._states = kwargs.get('STATES')
        self._register_type = kwargs.get('register_type')
        self.state = None
        self.addr = kwargs.get('ADDR')

    @property
    def name(self) -> str:
        """Get the name of the register."""
        return self._name

    @property
    def register_type(self) -> str:
        """Get the register type."""
        return self._register_type

    def get_id(self) -> str:
        """Get the ID of the register."""
        return self._register_id



"""Coil."""

from .baseregister import BaseRegister
from ..exceptions import PysmartyException


class Coil(BaseRegister):
    """Smart Coil Register."""

    def __init__(self, **kwargs):
        super().__init__(register_type='coil', **kwargs)

    def update_state(self):
        """Read Register."""
        res = self._connection.client.read_coils(
            self.addr, slave=self._connection.slave)
        if not res.isError():
            self.state = res.bits[0]

    def set_state(self, state):
        """Write Register."""
        res = self._connection.client.write_coil(
            self.addr, state, slave=self._connection.slave)
        if not res.isError():
            self.state = state



"""Modbus Client."""

from pymodbus.client import ModbusTcpClient


class Connection:
    """Modbus Client."""

    def __init__(self, host, port, slave=1, loop=None):
        """Modbus Connection init."""

        # pylint: disable=unsubscriptable-object
        self._client = ModbusTcpClient(host=host, port=port)
        self._slave = slave

    @property
    def client(self):
        """Get Modbus Client."""
        return self._client

    @property
    def host(self):
        """Get Host."""
        return self._client.host

    @property
    def port(self):
        """Get Port."""
        return self._client.port

    @property
    def slave(self):
        """Get Slave."""
        return self._slave

    def is_connected(self):
        """Return connection state. Attempt to connect if not already connected."""
        if not self._client.is_socket_open():
            self._client.connect()
        return self._client.is_socket_open()



"""Setup for pysmarty2 package."""

import setuptools


with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="pysmarty2",
    version="0.10.1",
    author="Martins Sipenko, Theo Nicolaum",
    author_email="martins.sipenko@gmail.com",
    description="Python API for Salda Smarty Modbus TCP",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/martinssipenko/pysmarty2",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=list(val.strip() for val in open('requirements.txt')),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Input Register."""

from .baseregister import BaseRegister


class InputRegister(BaseRegister):
    """Smarty Input Register."""

    def __init__(self, **kwargs):
        super().__init__(register_type='input_register', **kwargs)
        self.multiplier = kwargs.get('MULTIPLIER')
        self.unit_of_mesurement = kwargs.get('UNIT_OF_MESUREMENT')

    @property
    def value(self):
        """Register State With Multiplier."""
        return round(self.state * self.multiplier, 2) if self.state else self.state

    @property
    def state_name(self):
        """Register State Name."""
        return self._states.get(str(self.state))

    def update_state(self):
        """Read Register."""
        res = self._connection.client.read_input_registers(
            self.addr, slave=self._connection.slave)
        if not res.isError():
            self.state = res.registers[0]
